@@ -12,10 +12,8 @@ from scipy import interpolate
 from scipy.integrate import solve_ivp
 import warnings
 from collections import OrderedDict
-import functools
 
 from ..core.equilibrium import Equilibrium, EquilibriumRegion, Point2D
-from ..core.mesh import MultiLocationArray
 
 from ..utils import critical, polygons
 from ..utils.utils import with_default
@@ -53,6 +51,33 @@ class TokamakEquilibrium(Equilibrium):
             ),
             value_type=bool,
         ),
+        leg_refine_atol=WithMeta(
+            1.0e-5,
+            doc=(
+                "Tolerance used when iteratively refining the position of start points "
+                "of legs around an X-point."
+            ),
+            value_type=float,
+            check_all=is_positive,
+        ),
+        leg_refine_maxits=WithMeta(
+            1000,
+            doc=(
+                "Maximum number of iterations for iterative refinement of start points "
+                "of legs around an X-point."
+            ),
+            value_type=int,
+            check_all=is_positive,
+        ),
+        leg_trace_atol=WithMeta(
+            lambda options: 0.01 * options.refine_atol,
+            doc=(
+                "Tolerance used when calling scipy.integrate.solve_ivp to follow the "
+                "divertor legs. Default is 0.01*refine_atol."
+            ),
+            value_type=float,
+            check_all=is_positive,
+        ),
         nx_core=WithMeta(
             5,
             doc="Number of radial points in the core",
@@ -231,6 +256,11 @@ class TokamakEquilibrium(Equilibrium):
             ),
             value_type=[float, int, NoneType],
         ),
+        reverse_Bt=WithMeta(
+            False,
+            doc="Reverse the sign of toroidal magnetic field Bt.",
+            value_type=bool,
+        ),
         start_at_upper_outer=WithMeta(
             False,
             doc=(
@@ -257,6 +287,24 @@ class TokamakEquilibrium(Equilibrium):
             value_type=float,
             check_all=[is_positive, lambda x: x < 1.0],
         ),
+        xpoint_refine_atol=WithMeta(
+            1.0e-6,
+            doc=(
+                "Tolerance used when iteratively refining the position of an X-point "
+                "that has been identified."
+            ),
+            value_type=float,
+            check_all=is_positive,
+        ),
+        xpoint_refine_maxits=WithMeta(
+            1000,
+            doc=(
+                "Maximum number of iterations for iterative refinement of the position "
+                "of an X-point that has been identified."
+            ),
+            value_type=int,
+            check_all=is_positive,
+        ),
     )
 
     def __init__(
@@ -268,8 +316,10 @@ class TokamakEquilibrium(Equilibrium):
         fpol1D,
         pressure=None,
         wall=None,
+        psi_axis_gfile=None,
+        psi_bdry_gfile=None,
         psi_axis=None,
-        dct=False,
+        psi_bdry=None,
         make_regions=True,
         settings=None,
         nonorthogonal_settings=None,
@@ -294,12 +344,12 @@ class TokamakEquilibrium(Equilibrium):
         wall = [(R0,Z0), (R1, Z1), ...]
                A list of coordinate pairs, defining the vessel wall.
                The wall is closed, so the last point connects to the first.
-        psi_axis = float
-               The value of poloidal flux on the magnetic axis. If not
-               given, the value found at the axis will be used.
-        dct = bool
-               EXPERIMENTAL: If true, use a DCT to interpolate and differentiate
-               poloidal flux. By default a cubic spline is used.
+        psi_axis_gfile = float
+               The value of poloidal flux on the magnetic axis, given by the EFIT file.
+               psi_axis is the value calculated on the O-point.
+        psi_bdry_gfile = float
+               The value of poloidal flux at the plasma boundary, given by the EFIT file.
+               psi_bdry is the value calculated on the X-point.
         make_regions = bool
                Generate the regions to be meshed. The default (True)
                means that the object is complete after initialisation.
@@ -319,13 +369,18 @@ class TokamakEquilibrium(Equilibrium):
             psi2D *= -1.0
             psi1D *= -1.0
 
+        if self.user_options.reverse_Bt:
+            warnings.warn("Reversing the sign of the toroidal field")
+            fpol1D *= -1.0
+
+        self.psi_increasing = psi1D[-1] > psi1D[0]
+
         if self.user_options.extrapolate_profiles:
             # Extend the array to outer-most psi on grid
             if pressure is not None:
                 dpdpsi = (pressure[-1] - pressure[-2]) / (psi1D[-1] - psi1D[-2])
 
-            psi_increasing = psi1D[-1] > psi1D[0]
-            if psi_increasing:
+            if self.psi_increasing:
                 psi_outer = max(
                     self.user_options.psi_sol, self.user_options.psi_sol_inner
                 )
@@ -333,8 +388,8 @@ class TokamakEquilibrium(Equilibrium):
                 psi_outer = min(
                     self.user_options.psi_sol, self.user_options.psi_sol_inner
                 )
-            if (psi_increasing and psi_outer > psi1D[-1]) or (
-                not psi_increasing and psi_outer < psi1D[-1]
+            if (self.psi_increasing and psi_outer > psi1D[-1]) or (
+                not self.psi_increasing and psi_outer < psi1D[-1]
             ):
                 # if psi_outer is not beyond the last point of psi1D, no need to extend
                 # Exclude first point since duplicates last point in core
@@ -351,20 +406,9 @@ class TokamakEquilibrium(Equilibrium):
                 # p = p0 * exp( (psi - psi0) * dpdpsi / p0)
                 pressure = np.concatenate([pressure, p0 * np.exp(psiSOL * dpdpsi / p0)])
 
-        if dct:
-            # Create an interpolation
-            # This sets the functions
-            #   self.psi
-            #   self.f_R
-            #   self.f_Z
-            #   self.Bp_R
-            #   self.Bp_Z
-            #   self.d2psidR2
-            #   self.d2psidZ2
-            #   self.d2psidRdZ
-            self.magneticFunctionsFromGrid(R1D, Z1D, psi2D)
-        else:
-            self.psi_func = interpolate.RectBivariateSpline(R1D, Z1D, psi2D)
+        self.magneticFunctionsFromGrid(
+            R1D, Z1D, psi2D, self.user_options.psi_interpolation_method
+        )
 
         self.f_psi_sign = 1.0
         if len(fpol1D) > 0:
@@ -396,17 +440,43 @@ class TokamakEquilibrium(Equilibrium):
 
         # Find critical points (O- and X-points)
         R2D, Z2D = np.meshgrid(R1D, Z1D, indexing="ij")
-        opoints, xpoints = critical.find_critical(R2D, Z2D, psi2D)
+        opoints, xpoints = critical.find_critical(
+            R2D,
+            Z2D,
+            psi2D,
+            self.user_options.xpoint_refine_atol,
+            self.user_options.xpoint_refine_maxits,
+        )
+
         if len(opoints) == 0:
             warnings.warn("No O-points found in TokamakEquilibrium input")
         else:
-            if psi_axis is None:
-                psi_axis = opoints[0][2]  # Psi on magnetic axis
+            self.psi_axis = opoints[0][2]  # Psi on magnetic axis
             self.o_point = Point2D(opoints[0][0], opoints[0][1])
-        self.psi_axis = psi_axis
+            self.psi_axis_gfile = psi_axis_gfile
+            if (
+                psi_axis_gfile is not None
+                and abs(self.psi_axis - psi_axis_gfile) > 1.0e-3
+            ):
+                raise ValueError(
+                    f"psi_axis from the gfile ({psi_axis}) is different from psi "
+                    f"calculated on the O-point ({self.psi_axis})"
+                )
 
         if len(xpoints) == 0:
             warnings.warn("No X-points found in TokamakEquilibrium input")
+        else:
+            self.psi_bdry = xpoints[0][2]  # Psi on primary X-point
+            self.x_point = Point2D(xpoints[0][0], xpoints[0][1])
+            self.psi_bdry_gfile = psi_bdry_gfile
+            if (
+                psi_bdry_gfile is not None
+                and abs(self.psi_bdry - psi_bdry_gfile) > 1.0e-3
+            ):
+                raise ValueError(
+                    f"psi_bdry from the gfile ({psi_bdry}) is different from psi "
+                    f"calculated at the primary X-point ({self.psi_bdry})"
+                )
 
         self.x_points = [Point2D(r, z) for r, z, psi in xpoints]
         self.psi_sep = [psi for r, z, psi in xpoints]
@@ -445,6 +515,8 @@ class TokamakEquilibrium(Equilibrium):
 
         # Print the table of options
         print(self.user_options.as_table(), flush=True)
+        if not self.user_options.orthogonal:
+            print(self.nonorthogonal_options.as_table(), flush=True)
 
         if make_regions:
             # Create self.regions
@@ -471,7 +543,8 @@ class TokamakEquilibrium(Equilibrium):
         inds = np.nonzero((psivals[1:] - psi_sep) * (psivals[:-1] - psi_sep) < 0.0)[0]
 
         # Currently only handle standard X-points (no snowflakes)
-        assert len(inds) == 2
+        if len(inds) != 2:
+            raise ValueError("Found more than 2 divertor legs")
 
         # Divide-and-conquer to get a points on the leg
         # This goes into a list leg_points = [(r,z),..]
@@ -489,7 +562,8 @@ class TokamakEquilibrium(Equilibrium):
             s2 = 1.0
             psi1 = psivals[ind]  # s = 0
 
-            while s2 - s1 > 1e-5:
+            count = 0
+            while s2 - s1 > self.user_options.leg_refine_atol:
                 smid = 0.5 * (s1 + s2)
                 psi_mid = self.psi(r0 + smid * dr, z0 + smid * dz)
 
@@ -499,6 +573,14 @@ class TokamakEquilibrium(Equilibrium):
                 else:
                     psi1 = psi_mid
                     s1 = smid
+                count += 1
+                if count > self.user_options.leg_refine_maxits:
+                    raise ValueError(
+                        f"Failed to find leg start position. Error {s2 - s1} is "
+                        f"greater than tolerance "
+                        f"{self.user_options.leg_refine_atol} after "
+                        f"{self.user_options.leg_refine_maxits} iterations."
+                    )
             smid = 0.5 * (s1 + s2)
             r = r0 + smid * dr
             z = z0 + smid * dz
@@ -529,7 +611,13 @@ class TokamakEquilibrium(Equilibrium):
             pos = leg  # Starting position
             while True:
                 # Integrate a distance "step" along the leg
-                solve_result = solve_ivp(dpos_dl, (0.0, step), pos)
+                solve_result = solve_ivp(
+                    dpos_dl,
+                    (0.0, step),
+                    pos,
+                    rtol=0.0,
+                    atol=self.user_options.leg_trace_atol,
+                )
                 newpos = (solve_result.y[0][1], solve_result.y[1][1])
 
                 # Check if we have crossed the boundary
@@ -587,7 +675,8 @@ class TokamakEquilibrium(Equilibrium):
         - self.regions         OrderedDict of EquilibriumRegion objects
 
         """
-        assert self.psi_axis is not None
+        if self.psi_axis is None:
+            raise ValueError("psi_axis has not been set")
 
         # psi values
         def psinorm_to_psi(psinorm):
@@ -635,7 +724,8 @@ class TokamakEquilibrium(Equilibrium):
         )
 
         # Check that there are only one or two left
-        assert 0 < len(self.x_points) <= 2
+        if not (0 < len(self.x_points) <= 2):
+            raise ValueError(f"Found unsupported number of X-points: {self.x_points}")
 
         if len(self.x_points) == 1:
             # Generate the specifications for a lower or upper single null
@@ -670,7 +760,11 @@ class TokamakEquilibrium(Equilibrium):
                         psi_vals    1D array of poloidal flux values. Length 2*nx+1
         connections    List of connections between regions
         """
-        assert len(self.x_points) == 1
+        if len(self.x_points) != 1:
+            raise ValueError(
+                f"Single-null case must have 1 X-point, got {self.x_points}"
+            )
+
         # Single null. Could be lower or upper
 
         # Find lines along the legs from X-point to target
@@ -846,7 +940,10 @@ class TokamakEquilibrium(Equilibrium):
         connections    List of connections between regions
         """
 
-        assert len(self.x_points) == 2
+        if len(self.x_points) != 2:
+            raise ValueError(
+                f"Double-null case must have 2 X-points, got {self.x_points}"
+            )
 
         if self.x_points[0].Z < self.o_point.Z:
             # Lower double null
@@ -860,8 +957,16 @@ class TokamakEquilibrium(Equilibrium):
         lower_psi = self.psi_sep[lower_xpt_ind]
         upper_psi = self.psi_sep[upper_xpt_ind]
 
-        assert np.isclose(lower_psi, self.psi(*lower_x_point))
-        assert np.isclose(upper_psi, self.psi(*upper_x_point))
+        if not np.isclose(lower_psi, self.psi(*lower_x_point)):
+            raise ValueError(
+                f"psi-value for lower separatrix {lower_psi} does not match value at "
+                "lower X-point {self.psi(*lower_x_point)}"
+            )
+        if not np.isclose(upper_psi, self.psi(*upper_x_point)):
+            raise ValueError(
+                f"psi-value for upper separatrix {upper_psi} does not match value at "
+                "upper X-point {self.psi(*upper_x_point)}"
+            )
 
         # Find lines along the legs from X-point to target
         lower_legs = self.findLegs(lower_x_point)
@@ -952,6 +1057,8 @@ class TokamakEquilibrium(Equilibrium):
         if nx_inter_sep == 0:
             print("Generating a connected double null", flush=True)
 
+            self.double_null_type = "connected"
+
             # Only use psi of inner separatrix, not the outer one (if it is slightly
             # different)
             segments["upper_pf"]["psi_end"] = self.psi_sep[0]
@@ -1031,10 +1138,34 @@ class TokamakEquilibrium(Equilibrium):
                 ("outer_core", 1, "outer_lower_divertor", 1),
             ]
 
+            segments = self.segmentsWithPsivals(segments)
+
+            psi_check_sign = 1.0 if self.psi_increasing else -1.0
+            if (
+                psi_check_sign * segments["outer_sol"]["psi_vals"][1]
+                < psi_check_sign * self.psi_sep[1]
+            ):
+                raise ValueError(
+                    "Cannot create connected double-null grid - second X-point would "
+                    "be outside first gridded flux-surface in the outer SOL. Try "
+                    "increasing psi_spacing_separatrix_multiplier or setting "
+                    "nx_inter_sep>0."
+                )
+            if (
+                psi_check_sign * segments["inner_sol"]["psi_vals"][1]
+                < psi_check_sign * self.psi_sep[1]
+            ):
+                raise ValueError(
+                    "Cannot create connected double-null grid - second X-point would "
+                    "be outside first gridded flux-surface in the inner SOL. Try"
+                    "increasing psi_spacing_separatrix_multiplier or setting "
+                    "nx_inter_sep>0."
+                )
+
             return (
                 leg_regions,
                 core_regions,
-                self.segmentsWithPsivals(segments),
+                segments,
                 connections,
             )
 
@@ -1053,6 +1184,8 @@ class TokamakEquilibrium(Equilibrium):
             if self.x_points[0] == lower_x_point:
                 print("Lower double null", flush=True)
 
+                self.double_null_type = "lower"
+
                 inner_lower_segments = ["lower_pf", "near_sol", "inner_sol"]
                 outer_lower_segments = ["lower_pf", "near_sol", "outer_sol"]
 
@@ -1108,6 +1241,9 @@ class TokamakEquilibrium(Equilibrium):
 
             else:
                 print("Upper double null", flush=True)
+
+                self.double_null_type = "upper"
+
                 inner_lower_segments = ["lower_pf", "lower_pf2", "inner_sol"]
                 outer_lower_segments = ["lower_pf", "lower_pf2", "outer_sol"]
 
@@ -1407,6 +1543,8 @@ class TokamakEquilibrium(Equilibrium):
                 # The following arguments are passed through to PsiContour
                 points=region["points"],  # list of Point2D objects on the line
                 psival=region["psi"],
+                Rrange=(self.Rmin, self.Rmax),
+                Zrange=(self.Zmin, self.Zmax),
             )
 
             # Grids of psi values in each segment
@@ -1436,7 +1574,8 @@ class TokamakEquilibrium(Equilibrium):
                     # Determine if poloidal flux is increasing or decreasing with radius
                     sign = np.sign(self.psi_sep[0] - self.psi_axis)
 
-                    assert region["psi"] is not None
+                    if region["psi"] is None:
+                        raise ValueError("No psi values in region")
                     leg_psi = region["psi"]
                     eqreg.pressure = lambda psi: self.pressure(
                         leg_psi + sign * abs(psi - leg_psi)
@@ -1489,76 +1628,18 @@ class TokamakEquilibrium(Equilibrium):
             [(key, region_objects[key]) for key in ordering if key in region_objects]
         )
 
-    def handleMultiLocationArray(getResult):
-        @functools.wraps(getResult)
-        # Define a function which handles MultiLocationArray arguments
-        def handler(self, *args):
-            if isinstance(args[0], MultiLocationArray):
-                for arg in args[1:]:
-                    assert isinstance(arg, MultiLocationArray), (
-                        "if first arg is a MultiLocationArray, then others must be as "
-                        "well"
-                    )
-                result = MultiLocationArray(args[0].nx, args[0].ny)
-
-                if all(arg.centre is not None for arg in args):
-                    result.centre = getResult(self, *(arg.centre for arg in args))
-
-                if all(arg.xlow is not None for arg in args):
-                    result.xlow = getResult(self, *(arg.xlow for arg in args))
-
-                if all(arg.ylow is not None for arg in args):
-                    result.ylow = getResult(self, *(arg.ylow for arg in args))
-
-                if all(arg.corners is not None for arg in args):
-                    result.corners = getResult(self, *(arg.corners for arg in args))
-            else:
-                result = getResult(self, *args)
-            return result
-
-        return handler
-
-    @handleMultiLocationArray
-    def psi(self, R, Z):
-        "Return the poloidal flux at the given (R,Z) location"
-        return self.psi_func(R, Z, grid=False)
-
-    @handleMultiLocationArray
-    def f_R(self, R, Z):
-        """returns the R component of the vector Grad(psi)/|Grad(psi)|**2."""
-        dpsidR = self.psi_func(R, Z, dx=1, grid=False)
-        dpsidZ = self.psi_func(R, Z, dy=1, grid=False)
-        return dpsidR / (dpsidR ** 2 + dpsidZ ** 2)
-
-    @handleMultiLocationArray
-    def f_Z(self, R, Z):
-        """returns the Z component of the vector Grad(psi)/|Grad(psi)|**2."""
-        dpsidR = self.psi_func(R, Z, dx=1, grid=False)
-        dpsidZ = self.psi_func(R, Z, dy=1, grid=False)
-        return dpsidZ / (dpsidR ** 2 + dpsidZ ** 2)
-
-    @handleMultiLocationArray
-    def Bp_R(self, R, Z):
-        """returns the R component of the poloidal magnetic field."""
-        return self.psi_func(R, Z, dy=1, grid=False) / R
-
-    @handleMultiLocationArray
-    def Bp_Z(self, R, Z):
-        """returns the Z component of the poloidal magnetic field."""
-        return -self.psi_func(R, Z, dx=1, grid=False) / R
-
-    @handleMultiLocationArray
+    @Equilibrium.handleMultiLocationArray
     def fpol(self, psi):
         """poloidal current function,
         returns fpol such that B_toroidal = fpol/R"""
         return self.f_spl(psi * self.f_psi_sign)
 
-    @handleMultiLocationArray
+    @Equilibrium.handleMultiLocationArray
     def fpolprime(self, psi):
         """psi-derivative of fpol"""
         return self.fprime_spl(psi * self.f_psi_sign)
 
-    @handleMultiLocationArray
+    @Equilibrium.handleMultiLocationArray
     def pressure(self, psi):
         """Plasma pressure in Pascals"""
         if self.p_spl is None:
@@ -1598,11 +1679,11 @@ def read_geqdsk(
     data = geq_read(filehandle)
 
     # Range of psi normalises psi derivatives
-    psi_boundary = data["sibdry"]
-    psi_axis = data["simagx"]
+    psi_bdry_gfile = data["sibdry"]
+    psi_axis_gfile = data["simagx"]
 
     # 1D grid on which fpol is defined. Goes from normalised psi 0 to 1
-    psi1D = np.linspace(psi_axis, psi_boundary, data["nx"], endpoint=True)
+    psi1D = np.linspace(psi_axis_gfile, psi_bdry_gfile, data["nx"], endpoint=True)
 
     R1D = np.linspace(
         data["rleft"], data["rleft"] + data["rdim"], data["nx"], endpoint=True
@@ -1632,6 +1713,8 @@ def read_geqdsk(
         psi2D,
         psi1D,
         fpol,
+        psi_bdry_gfile=psi_bdry_gfile,
+        psi_axis_gfile=psi_axis_gfile,
         pressure=pressure,
         wall=wall,
         make_regions=make_regions,
