@@ -123,6 +123,21 @@ class TORPEXMagneticField(Equilibrium):
         ),
         saddle_point_p1=[0.85, -0.15],
         saddle_point_p2=[0.85, 0.15],
+        # Tolerance for positioning points that should be at X-point, but need to be
+        # slightly displaced from the null so code can follow Grad(psi).
+        # Number between 0. and 1.
+        xpoint_offset=WithMeta(
+            0.1,
+            doc=(
+                "Tolerance for placing intial positions for tracing perpendiculars "
+                "that should start exactly at an X-point, but initial positions to be "
+                "slightly displaced from the null so code can follow Grad(psi).  This "
+                "is a numerical fudge factor that may need to be increased for "
+                "low-resolution input equilibria."
+            ),
+            value_type=float,
+            check_all=[is_positive, lambda x: x < 1.0],
+        ),
     )
 
     def __init__(self, equilibOptions, meshOptions):
@@ -137,15 +152,17 @@ class TORPEXMagneticField(Equilibrium):
 
         print(self.user_options.as_table())
 
-        # Call Equilibrium constructor after adding stuff to options
-        super().__init__(meshOptions)
-
         if "Coils" in equilibOptions:
             self.coils = [Coil(**c) for c in equilibOptions["Coils"]]
 
             self.magneticFunctionsFromCoils()
 
             self.Bt_axis = equilibOptions["Bt_axis"]
+
+            self.Rmin = -float("inf")
+            self.Rmax = float("inf")
+            self.Zmin = -float("inf")
+            self.Zmax = float("inf")
         elif "gfile" in equilibOptions:
             # load a g-file
             with open(equilibOptions["gfile"], "rt") as fh:
@@ -159,10 +176,11 @@ class TORPEXMagneticField(Equilibrium):
                 gfile["zmid"] + 0.5 * gfile["zdim"],
                 gfile["ny"],
             )
-            # Note we expect first index (row) of psirz to change with the height Z, and
-            # the second (column) to change with major radius R, which is the opposite
-            # from _geqdsk's (from FreeGS) convention, so transpose
-            psirz = gfile["psi"].T
+            self.Rmin = R[0]
+            self.Rmax = R[-1]
+            self.Zmin = Z[0]
+            self.Zmax = Z[-1]
+            psirz = gfile["psi"]
 
             # check sign of psirz is consistent with signs of psi_axis, psi_bndry and
             # plasma current
@@ -176,17 +194,19 @@ class TORPEXMagneticField(Equilibrium):
                 # plane (outward in major radius at the top of the torus, inward at the
                 # bottom).  This corresponds to plasma current in the anti-clockwise
                 # toroidal direction looking from above, so current should be positive
-                assert Ip >= 0.0, (
-                    "direction of plasma current should be anti-clockwise to be "
-                    "consistent with sign of grad(psi)"
-                )
+                if Ip < 0.0:
+                    raise ValueError(
+                        "direction of plasma current should be anti-clockwise to be "
+                        "consistent with sign of grad(psi)"
+                    )
             else:
                 # psi decreases outward radially, so current should be in the opposite
                 # direction
-                assert Ip <= 0.0, (
-                    "direction of plasma current should be clockwise to be consistent "
-                    "with sign of grad(psi)"
-                )
+                if Ip > 0.0:
+                    raise ValueError(
+                        "direction of plasma current should be clockwise to be "
+                        "consistent with sign of grad(psi)"
+                    )
             # index of a point close to the magnetic axis
             i_axis = numpy.searchsorted(R, R_axis)
             j_axis = numpy.searchsorted(Z, Z_axis)
@@ -224,7 +244,9 @@ class TORPEXMagneticField(Equilibrium):
                     # axis. Should be going towards psi_bndry. So flip the sign
                     psirz = -psirz
 
-            self.magneticFunctionsFromGrid(R, Z, psirz)
+            self.magneticFunctionsFromGrid(
+                R, Z, psirz, self.user_options.psi_interpolation_method
+            )
 
             self.Bt_axis = gfile["bcentr"]
         elif "matfile" in equilibOptions:
@@ -248,16 +270,30 @@ class TORPEXMagneticField(Equilibrium):
             R = R[Zinds, :]
             Z = Z[:, Rinds]
             Z = Z[Zinds, :]
+            self.Rmin = R[0, 0]
+            self.Rmax = R[0, -1]
+            self.Zmin = Z[0, 0]
+            self.Zmax = Z[-1, 0]
             psi = psi[:, Rinds]
             psi = psi[Zinds, :]
 
-            self.magneticFunctionsFromGrid(R[0, :], Z[:, 0], psi)
+            self.magneticFunctionsFromGrid(
+                R[0, :], Z[:, 0], psi.T, self.user_options.psi_interpolation_method
+            )
 
             Bt = eqfile["Bphi"][0, 0]
             RindMid = Bt.shape[1] // 2
             ZindMid = Bt.shape[0] // 2
-            assert eqfile["R"][0, 0][ZindMid, RindMid] == 1.0
-            assert eqfile["Z"][0, 0][ZindMid, RindMid] == 0.0
+            if eqfile["R"][0, 0][ZindMid, RindMid] != 1.0:
+                raise ValueError(
+                    f"R at centre should be 1.0, got "
+                    f"{eqfile['R'][0, 0][ZindMid, RindMid]}"
+                )
+            if eqfile["Z"][0, 0][ZindMid, RindMid] != 0.0:
+                raise ValueError(
+                    f"Z at centre should be 0.0, got "
+                    f"{eqfile['Z'][0, 0][ZindMid, RindMid]}"
+                )
             self.Bt_axis = Bt[ZindMid, RindMid]
         else:
             raise ValueError("Failed to initialise psi function from inputs")
@@ -285,6 +321,10 @@ class TORPEXMagneticField(Equilibrium):
                 "Warning: failed to find X-point. Equilibrium generation will fail"
             )
 
+        # Call Equilibrium constructor after adding stuff to options
+        super().__init__(meshOptions)
+        print(self.nonorthogonal_options.as_table())
+
     def TORPEX_wall(self, theta):
         """
         Return the location of the TORPEX wall parameterized by the angle theta
@@ -460,7 +500,10 @@ class TORPEXMagneticField(Equilibrium):
         For TORPEX, follow 4 legs away from the x-point, starting with a rough guess and
         then refining to the separatrix value of A_toroidal.
         """
-        assert len(self.x_points) == 1, "should be one X-point for TORPEX configuration"
+        if len(self.x_points) != 1:
+            raise ValueError(
+                f"should be one X-point for TORPEX configuration, got {self.x_points}"
+            )
         xpoint = self.x_points[0]
 
         boundary = self.findRoots_1d(
@@ -492,7 +535,7 @@ class TORPEXMagneticField(Equilibrium):
 
         self.regions = OrderedDict()
         wall_vectors = OrderedDict()
-        s = numpy.linspace(10.0 * self.user_options.refine_atol, 1.0, npoints)
+        s = numpy.linspace(self.user_options.xpoint_offset / npoints, 1.0, npoints)
         for i, boundary_position in enumerate(boundary):
             name = legnames[i]
             boundaryPoint = self.wallPosition(boundary_position)
@@ -502,16 +545,16 @@ class TORPEXMagneticField(Equilibrium):
                 equilibrium=self,
                 name=name,
                 nSegments=2,
-                settings=dict(self.user_options),
-                nonorthogonal_settings=dict(self.nonorthogonal_options),
                 nx=legoptions[name]["nx"],
                 ny=legoptions[name]["ny"],
                 kind=legoptions[name]["kind"],
                 ny_total=self.ny_total,
                 points=[Point2D(R, Z) for R, Z in zip(legR, legZ)],
                 psival=self.psi_sep[0],
+                Rrange=(self.Rmin, self.Rmax),
+                Zrange=(self.Zmin, self.Zmax),
             )
-            self.regions[name] = leg.getRefined()
+            self.regions[name] = leg.getRefined(psi=self.psi)
             wall_vectors[name] = self.wallVector(boundary_position)
 
         # Make the SeparatrixContours go around clockwise
@@ -635,11 +678,14 @@ def createMesh(filename):
 
     equilibrium = TORPEXMagneticField(equilibOptions, meshOptions)
 
+    # Pick up any changes due to default options, etc.
+    meshOptions.update(equilibrium.user_options)
+
     print("X-point", equilibrium.x_points[0], "with psi=" + str(equilibrium.psi_sep[0]))
 
     equilibrium.makeRegions()
 
-    return BoutMesh(equilibrium, settings=equilibrium.user_options)
+    return BoutMesh(equilibrium, settings=meshOptions)
 
 
 def createEqdsk(equilib, *, nR, Rmin, Rmax, nZ, Zmin, Zmax, filename="torpex_test.g"):
