@@ -8,10 +8,22 @@
 #
 #
 
+import gc
 import warnings
 
 
-def main():
+def main(*, add_noise=None):
+    """
+    Read a g-file and (optional) input file, and write a grid file
+
+    Parameters
+    ----------
+    add_noise : int, optional
+        Only intended for use in tests. If an int is passed, used as the seed for a
+        random number generator adding small amounts of noise to the EquilibriumRegion
+        points before generating the grid.
+    """
+
     from argparse import ArgumentParser
 
     parser = ArgumentParser()
@@ -55,6 +67,20 @@ def main():
     with open(filename, "rt") as fh:
         eq = tokamak.read_geqdsk(fh, settings=options, nonorthogonal_settings=options)
 
+    if add_noise is not None:
+        # Add machine-precision level noise for testing robustness of grid generation
+        import numpy as np
+        from ..core.equilibrium import Point2D
+
+        rng = np.random.default_rng(add_noise)
+        for region in eq.regions.values():
+            region.points = [
+                Point2D(
+                    p.R + rng.normal(scale=1.0e-16), p.Z + rng.normal(scale=5.0e-17)
+                )
+                for p in region.points
+            ]
+
     if options.get("plot_regions", False):
         try:
             import matplotlib.pyplot as plt
@@ -62,7 +88,9 @@ def main():
             eq.plotPotential(ncontours=40)
             for region in eq.regions.values():
                 plt.plot(
-                    [p.R for p in region.points], [p.Z for p in region.points], "-o"
+                    [p.R for p in region.points],
+                    [p.Z for p in region.points],
+                    marker="o",
                 )
             print("Close window to continue...")
             plt.show()
@@ -94,6 +122,15 @@ def main():
 
     mesh.writeGridfile(options.get("grid_file", "bout.grd.nc"))
 
+    # Delete and garbage-collect hypnotoad objects here so that any ParallelMap
+    # instances get deleted if they exists. ParallelMap.__del__() calls
+    # terminate on the worker processes. Needs to happen before program exits
+    # or program will hang waiting for parallel workers to finish (which they
+    # never do because they are waiting for another job in
+    # ParallelMap.worker_run()).
+    del eq, mesh
+    gc.collect()
+
 
 if __name__ == "__main__":
     main()
